from typing import Final
from . import Device


class EnergyManagerDevice:
    """The EnergyManager device itself. Needed to get the device model and serial number."""

    DEVICE_CLASS: Final = "com.kiwigrid.devices.em.EnergyManager"

    # Tag name constants
    TAG_MODEL: Final[str] = "IdModelCode"
    TAG_FIRMWARE: Final[str] = "IdFirmware"

    def __init__(self, device: Device):
        """Wrap the device with EnergyManagerDevice."""
        self.device = device

    @property
    def model(self) -> str:
        """Get the EnergyManager model."""
        return self.device.get_tag_value_as_str(EnergyManagerDevice.TAG_MODEL)

    @property
    def firmware(self) -> str:
        """Get the EnergyManager firmware."""
        return self.device.get_tag_value_as_str(EnergyManagerDevice.TAG_FIRMWARE)




"""Code to query and parse data from the EnergyManager."""

import aiohttp
import logging
from typing import Optional
from . import EnergyManagerData

_LOGGER = logging.getLogger(__name__)

class EnergyManager:
    """The energy manager accessor class."""

    _URL = ""

    def __init__(self, host):
        """Iinitialize class, generate URL."""
        self.host = host
        self._URL = f"http://{self.host}/rest/kiwigrid/wizard/devices"

    async def test_connection(self) -> str:
        """Test the connection. Return value is the EnergyManager serial number."""
        try:
            json_data = await self.query_json_data()
        except Exception:
            raise CannotConnect

        data = EnergyManagerData(json_data)
        energy_manager = data.energy_manager_device
        try:
            if energy_manager is None:
                raise CannotParseData
        except Exception:
            raise CannotParseData
        return energy_manager.device.guid

    async def get_data(self) -> Optional[EnergyManagerData]:
        """Query the EnergyManager and return the parsed data."""
        try:
            json = await self.query_json_data()
        except Exception as e:
            _LOGGER.error("Error querying EnergyManager at %s: %s", self._URL, e)
            return None

        try:
            devices = EnergyManagerData(json)
            return devices
        except Exception as e:
            _LOGGER.error("Error parsing JSON data from EnergyManager: %s", e)
            return None

    async def query_json_data(self):
        """Query the JSON data from the EnergyManager."""
        _LOGGER.debug("Querying EnergyManager at %s", self._URL)
        async with aiohttp.ClientSession() as session:
            async with session.get(self._URL) as response:
                _LOGGER.debug("Received response status code %s", response.status)
                return await response.json()


class CannotConnect(Exception):
    """Error to indicate we cannot connect."""


class CannotParseData(Exception):
    """Error to indicate that the data cannot be parsed."""



from typing import Any, Dict, List


class Device:
    """The base definition for an EnergyManager device."""

    # The unique ID of the device.
    guid: str = ""

    # The device classes defined for this device.
    device_classes: List[str] = []

    # The tag items of the device.
    tag_items: Dict[str, Any] = {}

    def read_json_item(self, item):
        """Read the item JSON data into this object."""
        self.guid = item["guid"] if "guid" in item else ""
        self.device_classes = Device.get_item_device_classes(item)
        self.tag_items = Device.get_tag_items(item)

    def get_tag_value_as_str(self, tag_name) -> str:
        """Get the item tag value as a string."""
        try:
            if tag_name in self.tag_items:
                return str(self.tag_items[tag_name])
            return ""
        except Exception:
            return ""

    def get_tag_value_as_int(self, tag_name) -> int:
        """Get the item tag value as an int."""
        try:
            return int(float(self.tag_items[tag_name]))
        except Exception:
            return int(0)

    def get_tag_value_as_float(self, tag_name) -> float:
        """Get the item tag value as a float."""
        try:
            return float(self.tag_items[tag_name])
        except Exception:
            return float(0)

    @staticmethod
    def get_item_device_classes(item: dict) -> List[str]:
        """Get the devices classes of the items."""
        classes: List[str] = []
        if "deviceModel" in item:
            device_model_items = item["deviceModel"]
            for device_model_item in device_model_items:
                if "deviceClass" in device_model_item:
                    classes.append(device_model_item["deviceClass"])
        return classes

    @staticmethod
    def get_tag_items(item: dict) -> Dict[str, Any]:
        """Get the tag items of the item."""
        items: Dict[str, Any] = {}
        if "tagValues" in item:
            tag_values = item["tagValues"]
            for tag_name, tag_value in tag_values.items():
                items[tag_name] = tag_value["value"] if "value" in tag_value else ""
        return items




from typing import Final
from . import Device

class LocationDevice:
    """
    The Location EnergyManager device. Contains most of the interesting info.

    The power tags contain the instantaneous power readings in W.
    The work tags contain the accumulated energy values in Wh.
    """

    DEVICE_CLASS: Final[str] = "com.kiwigrid.devices.location.Location"

    # Tag name constants
    TAG_POWER_BUFFERED: Final[str] = "PowerBuffered"
    TAG_POWER_BUFFERED_FROM_GRID: Final[str] = "PowerBufferedFromGrid"
    TAG_POWER_BUFFERED_FROM_PRODUCERS: Final[str] = "PowerBufferedFromProducers"
    TAG_POWER_CONSUMED: Final[str] = "PowerConsumed"
    TAG_POWER_CONSUMED_FROM_GRID: Final[str] = "PowerConsumedFromGrid"
    TAG_POWER_CONSUMED_FROM_PRODUCERS: Final[str] = "PowerConsumedFromProducers"
    TAG_POWER_CONSUMED_FROM_STORAGE: Final[str] = "PowerConsumedFromStorage"
    TAG_POWER_IN: Final[str] = "PowerIn"
    TAG_POWER_OUT: Final[str] = "PowerOut"
    TAG_POWER_OUT_FROM_PRODUCERS: Final[str] = "PowerOutFromProducers"
    TAG_POWER_OUT_FROM_STORAGE: Final[str] = "PowerOutFromStorage"
    TAG_POWER_PRODUCED: Final[str] = "PowerProduced"
    TAG_POWER_RELEASED: Final[str] = "PowerReleased"
    TAG_POWER_SELF_CONSUMED: Final[str] = "PowerSelfConsumed"
    TAG_POWER_SELF_SUPPLIED: Final[str] = "PowerSelfSupplied"

    TAG_WORK_BUFFERED: Final[str] = "WorkBuffered"
    TAG_WORK_BUFFERED_FROM_GRID: Final[str] = "WorkBufferedFromGrid"
    TAG_WORK_BUFFERED_FROM_PRODUCERS: Final[str] = "WorkBufferedFromProducers"
    TAG_WORK_CONSUMED: Final[str] = "WorkConsumed"
    TAG_WORK_CONSUMED_FROM_GRID: Final[str] = "WorkConsumedFromGrid"
    TAG_WORK_CONSUMED_FROM_PRODUCERS: Final[str] = "WorkConsumedFromProducers"
    TAG_WORK_CONSUMED_FROM_STORAGE: Final[str] = "WorkConsumedFromStorage"
    TAG_WORK_IN: Final[str] = "WorkIn"
    TAG_WORK_OUT: Final[str] = "WorkOut"
    TAG_WORK_OUT_FROM_PRODUCERS: Final[str] = "WorkOutFromProducers"
    TAG_WORK_OUT_FROM_STORAGE: Final[str] = "WorkOutFromStorage"
    TAG_WORK_PRODUCED: Final[str] = "WorkProduced"
    TAG_WORK_RELEASED: Final[str] = "WorkReleased"
    TAG_WORK_SELF_CONSUMED: Final[str] = "WorkSelfConsumed"
    TAG_WORK_SELF_SUPPLIED: Final[str] = "WorkSelfSupplied"

    def __init__(self, device: Device):
        """Wrapper for LocationDevice."""
        self.device = device

    @property
    def power_buffered(self) -> float:
        """The power being stored to battery, in W."""
        return self.device.get_tag_value_as_float(LocationDevice.TAG_POWER_BUFFERED)

    @property
    def power_buffered_from_grid(self) -> float:
        """The power being stored to battery from the grid, in W."""
        return self.device.get_tag_value_as_float(LocationDevice.TAG_POWER_BUFFERED_FROM_GRID)

    @property
    def power_buffered_from_producers(self) -> float:
        """The power being stored to battery from the producers (eg - solar), in W."""
        return self.device.get_tag_value_as_float(LocationDevice.TAG_POWER_BUFFERED_FROM_PRODUCERS)

    @property
    def power_consumed(self) -> float:
        """The power being consumed, in W."""
        return self.device.get_tag_value_as_float(LocationDevice.TAG_POWER_CONSUMED)

    @property
    def power_consumed_from_grid(self) -> float:
        """The power being consumed from the grid, in W."""
        return self.device.get_tag_value_as_float(LocationDevice.TAG_POWER_CONSUMED_FROM_GRID)

    @property
    def power_consumed_from_producers(self) -> float:
        """The power being consumed from producers (eg - solar), in W."""
        return self.device.get_tag_value_as_float(LocationDevice.TAG_POWER_CONSUMED_FROM_PRODUCERS)

    @property
    def power_consumed_from_storage(self) -> float:
        """The power being consumed from battery, in W."""
        return self.device.get_tag_value_as_float(LocationDevice.TAG_POWER_CONSUMED_FROM_STORAGE)

    @property
    def power_in(self) -> float:
        """The power coming in from the grid, in W."""
        return self.device.get_tag_value_as_float(LocationDevice.TAG_POWER_IN)

    @property
    def power_out(self) -> float:
        """The power going out to the grid, in W."""
        return self.device.get_tag_value_as_float(LocationDevice.TAG_POWER_OUT)

    @property
    def power_out_from_producers(self) -> float:
        """The power going out to the grid from producers (eg - solar), in W."""
        return self.device.get_tag_value_as_float(LocationDevice.TAG_POWER_OUT_FROM_PRODUCERS)

    @property
    def power_out_from_storage(self) -> float:
        """The power going out to the grid from battery, in W."""
        return self.device.get_tag_value_as_float(LocationDevice.TAG_POWER_OUT_FROM_STORAGE)

    @property
    def power_produced(self) -> float:
        """The power produced (eg - solar), in W."""
        return self.device.get_tag_value_as_float(LocationDevice.TAG_POWER_PRODUCED)

    @property
    def power_released(self) -> float:
        """The power relased from the battery, in W."""
        return self.device.get_tag_value_as_float(LocationDevice.TAG_POWER_RELEASED)

    @property
    def power_self_consumed(self) -> float:
        """The power self consumed from producers, in W."""
        return self.device.get_tag_value_as_float(LocationDevice.TAG_POWER_SELF_CONSUMED)

    @property
    def power_self_supplied(self) -> float:
        """The power self supplied (to storage?), in W."""
        return self.device.get_tag_value_as_float(LocationDevice.TAG_POWER_SELF_SUPPLIED)

    @property
    def work_buffered(self) -> float:
        """The energy being stored to battery, in Wh."""
        return self.device.get_tag_value_as_float(LocationDevice.TAG_WORK_BUFFERED)

    @property
    def work_buffered_from_grid(self) -> float:
        """The energy being stored to battery from the grid, in Wh."""
        return self.device.get_tag_value_as_float(LocationDevice.TAG_WORK_BUFFERED_FROM_GRID)

    @property
    def work_buffered_from_producers(self) -> float:
        """The energy being stored to battery from the producers (eg - solar), in Wh."""
        return self.device.get_tag_value_as_float(LocationDevice.TAG_WORK_BUFFERED_FROM_PRODUCERS)

    @property
    def work_consumed(self) -> float:
        """The energy being consumed, in Wh."""
        return self.device.get_tag_value_as_float(LocationDevice.TAG_WORK_CONSUMED)

    @property
    def work_consumed_from_grid(self) -> float:
        """The energy being consumed from the grid, in Wh."""
        return self.device.get_tag_value_as_float(LocationDevice.TAG_WORK_CONSUMED_FROM_GRID)

    @property
    def work_consumed_from_producers(self) -> float:
        """The energy being consumed from producers (eg - solar), in Wh."""
        return self.device.get_tag_value_as_float(LocationDevice.TAG_WORK_CONSUMED_FROM_PRODUCERS)

    @property
    def work_consumed_from_storage(self) -> float:
        """The energy being consumed from battery, in Wh."""
        return self.device.get_tag_value_as_float(LocationDevice.TAG_WORK_CONSUMED_FROM_STORAGE)

    @property
    def work_in(self) -> float:
        """The energy coming in from the grid, in Wh."""
        return self.device.get_tag_value_as_float(LocationDevice.TAG_WORK_IN)

    @property
    def work_out(self) -> float:
        """The energy going out to the grid, in Wh."""
        return self.device.get_tag_value_as_float(LocationDevice.TAG_WORK_OUT)

    @property
    def work_out_from_producers(self) -> float:
        """The energy going out to the grid from producers (eg - solar), in Wh."""
        return self.device.get_tag_value_as_float(LocationDevice.TAG_WORK_OUT_FROM_PRODUCERS)

    @property
    def work_out_from_storage(self) -> float:
        """The energy going out to the grid from battery, in Wh."""
        return self.device.get_tag_value_as_float(LocationDevice.TAG_WORK_OUT_FROM_STORAGE)

    @property
    def work_produced(self) -> float:
        """The energy produced (eg - solar), in Wh."""
        return self.device.get_tag_value_as_float(LocationDevice.TAG_WORK_PRODUCED)

    @property
    def work_released(self) -> float:
        """The energy relased from the battery, in Wh."""
        return self.device.get_tag_value_as_float(LocationDevice.TAG_WORK_RELEASED)

    @property
    def work_self_consumed(self) -> float:
        """The energy self consumed from producers, in Wh."""
        return self.device.get_tag_value_as_float(LocationDevice.TAG_WORK_SELF_CONSUMED)

    @property
    def work_self_supplied(self) -> float:
        """The energy self supplied (to storage?), in Wh."""
        return self.device.get_tag_value_as_float(LocationDevice.TAG_WORK_SELF_SUPPLIED)



from typing import Final
from . import Device

class BatteryConverterDevice:
    """The battery converter EnergyManager device. Contains info about flow to and from battery."""

    DEVICE_CLASS: Final = "com.kiwigrid.devices.batteryconverter.BatteryConverter"

    # Tag name constants
    TAG_CURRENT_BATTERY_IN: Final[str] = "CurrentBatteryIn"
    TAG_CURRENT_BATTERY_OUT: Final[str] = "CurrentBatteryOut"
    TAG_CURRENT_GRM_IN: Final[str] = "CurrentGRMIn"
    TAG_CURRENT_GRM_OUT: Final[str] = "CurrentGRMOut"
    TAG_CURRENT_STRING_DC_IN: Final[str] = "CurrentStringDCIn"
    TAG_MODE_CONVERTER: Final[str] = "ModeConverter"
    TAG_POWER_AC_IN: Final[str] = "PowerACIn"
    TAG_POWER_AC_OUT: Final[str] = "PowerACOut"
    TAG_STATE_OF_CHARGE: Final[str] = "StateOfCharge"
    TAG_STATE_OF_HEALTH: Final[str] = "StateOfHealth"
    TAG_TEMPERATURE_BATTERY: Final[str] = "TemperatureBattery"
    TAG_VOLTAGE_GRM_IN: Final[str] = "VoltageGRMIn"
    TAG_VOLTAGE_GRM_OUT: Final[str] = "VoltageGRMOut"
    TAG_VOLTAGE_BATTERY_CELL_MEAN: Final[str] = "VoltageBatteryCellMean"
    TAG_VOLTAGE_BATTERY_STRING: Final[str] = "VoltageBatteryString"
    TAG_WORK_AC_IN: Final[str] = "WorkACIn"
    TAG_WORK_AC_OUT: Final[str] = "WorkACOut"

    def __init__(self, device: Device):
        """Wrap the device with BatteryConverterDevice."""
        self.device = device

    @property
    def current_battery_in(self) -> float:
        """The battery input current, in A."""
        return self.device.get_tag_value_as_float(BatteryConverterDevice.TAG_CURRENT_BATTERY_IN)

    @property
    def current_battery_out(self) -> float:
        """The battery output current, in A."""
        return self.device.get_tag_value_as_float(BatteryConverterDevice.TAG_CURRENT_BATTERY_OUT)

    @property
    def current_grm_in(self) -> float:
        """The GRM input current, in A."""
        return self.device.get_tag_value_as_float(BatteryConverterDevice.TAG_CURRENT_GRM_IN)

    @property
    def current_grm_out(self) -> float:
        """The GRM output current, in A."""
        return self.device.get_tag_value_as_float(BatteryConverterDevice.TAG_CURRENT_GRM_OUT)

    @property
    def current_dc_string_in(self) -> float:
        """The battery DC input current, in A."""
        return self.device.get_tag_value_as_float(BatteryConverterDevice.TAG_CURRENT_STRING_DC_IN)

    @property
    def mode_converter(self) -> str:
        """The current charging mode: CHARGING, ???"""
        return self.device.get_tag_value_as_str(BatteryConverterDevice.TAG_MODE_CONVERTER)

    @property
    def power_ac_in(self) -> float:
        """The input AC power, in W."""
        return self.device.get_tag_value_as_float(BatteryConverterDevice.TAG_POWER_AC_IN)

    @property
    def power_ac_out(self) -> float:
        """The output AC power, in W."""
        return self.device.get_tag_value_as_float(BatteryConverterDevice.TAG_POWER_AC_OUT)

    @property
    def state_of_charge(self) -> int:
        """The current battery charge state, in %."""
        return self.device.get_tag_value_as_int(BatteryConverterDevice.TAG_STATE_OF_CHARGE)

    @property
    def state_of_health(self) -> int:
        """The current battery health, in %."""
        return self.device.get_tag_value_as_int(BatteryConverterDevice.TAG_STATE_OF_HEALTH)

    @property
    def temperature_battery(self) -> int:
        """The current battery temperature, in C."""
        return self.device.get_tag_value_as_int(BatteryConverterDevice.TAG_TEMPERATURE_BATTERY)

    @property
    def voltage_grm_in(self) -> float:
        """The GRM input voltage, in V."""
        return self.device.get_tag_value_as_float(BatteryConverterDevice.TAG_VOLTAGE_GRM_IN)

    @property
    def voltage_grm_out(self) -> float:
        """The GRM output voltage, in V."""
        return self.device.get_tag_value_as_float(BatteryConverterDevice.TAG_VOLTAGE_GRM_OUT)

    @property
    def voltage_battery_cell_mean(self) -> float:
        """The mean battery cell voltage, in V."""
        return self.device.get_tag_value_as_float(BatteryConverterDevice.TAG_VOLTAGE_BATTERY_CELL_MEAN)

    @property
    def voltage_battery_cell_string(self) -> float:
        """The battery string voltage, in V."""
        return self.device.get_tag_value_as_float(BatteryConverterDevice.TAG_VOLTAGE_BATTERY_STRING)

    @property
    def work_ac_in(self) -> float:
        """The accumulated amount of AC energy received, in Wh."""
        return self.device.get_tag_value_as_float(BatteryConverterDevice.TAG_WORK_AC_IN)

    @property
    def work_ac_out(self) -> float:
        """The accumulated amount of AC energy sent, in Wh."""
        return self.device.get_tag_value_as_float(BatteryConverterDevice.TAG_WORK_AC_OUT)
